/*
 * Copyright 1994 Charles Blake and Michael K. Johnson
 * This file is a part of procps, which is distributable
 * under the conditions of the GNU Library General Public License.
 * See the file COPYING for details.
 */

#include <string.h>		/* for strcmp */
#include <stdio.h>		/* for parse error output */

#include "proc/procps.h"
#include "proc/readproc.h"	/* for proc_t */

/*
  This module was written by Charles Blake for procps.

mult_lvl_cmp:
    slick general purpose multi-level compare function I invented.
sort_depth:
    the number of levels of functions *to use*.  This means many more levels
    can be defined than mult_lvl_cmp tries out.  If this is 1 then mult_lvl_cmp
    is just a trivial wrapper around (*sort_function[0]).
sort_direction:
    multiplicative factor for the output of cmp_whatever.
    1 ==> default order, -1 ==> reverse order, 0 ==> forced equality
    The 0 bit is the neat part.  Since a value of zero is the code for equality
    multiplying the output of cmp_foo(a,b) forces a==b to be true.  This is a
    convenient way to turn sorting off in middle levels of a multi-level sort.
    If time is a problem, reforming the whole sort_function array to not include
    these unsorted middle levels will be faster since then cmp_foo won't even
    be called.  It might simplify some code depending upon how you organize it.
sort_function[]:
    array of function pointers that points to our family of comparison functions
    (I have named them cmp_* but mult_lvl_cmp doesn't care what they're named).
    This may be declared and initialized like so:
       int (*sort_function[])(void* a, void* b)={&cmp_foo, &cmp_bar, &cmp_hiho};
    You could also use my command line '-O' parser below.

Note that we only descend levels until the order is determined.  If we descend
all levels, that means that the items are equal at all levels, so we return 0.
Otherwise we return whatever the level's cmp_foo function would have returned.
This allows whatever default behavior you want for cmp_foo.  sort_direction[]
reverses this default behavior, but mult_lvl_cmp doesn't decide that ascending
or descending is the default.  That is the job of your cmp_foo's.
*/

/* the only reason these are global is because qsort(3) likes it that way.
   It's also a little more efficient if mult_lvl_cmp() is called many times.
*/

struct tree_node {
	proc_t *proc;
	pid_t pid;
	pid_t ppid;
	char *line;
	char *cmd;
	char **cmdline;
	char **environ;
	int children;
	int maxchildren;
	int *child;
	int have_parent;
};

int sort_depth = 0;
int sort_direction[10];		/* storage for 10 levels, but 4 would be plenty! */
int (*sort_function[10]) (void *a, void *b);

int mult_lvl_cmp(void *a, void *b)
{
	int i, cmp_val;
	for (i = 0; i < sort_depth; i++) {
		cmp_val = sort_direction[i] * (*sort_function[i]) (a, b);
		if (cmp_val != 0)
			return cmp_val;
	}
	return 0;
}

/* qsort(3) compliant comparison functions for all members of the ps_proc
   structure (in the same order in which they appear in the proc_t declaration)
   return is {-1,0,1} as {a<b, a==b, a>b}
   default ordering is ascending for all members. (flip 1,-1 to reverse)
*/

void reset_sort_options(void)
{
	int i;

	sort_depth = 0;
	for (i = 0; i < 10; i++) {
		sort_direction[i] = 0;
		sort_function[i] = (cmp_t) NULL;
	}
}

void register_sort_function(int dir, cmp_t func)
{
	sort_function[sort_depth] = func;
	sort_direction[sort_depth++] = dir;
}
