/*
 * Copyright (C) 1996 Charles Blake
 * Copyright (C) 1998 Michael K. Johnson
 * kernel address -> symbol with next lower address.  Charles Blake, 1996.
 * Written to obviate the need for psdatabase initialization based upon kernel
 * binary formats, etc.
 *
 * The basic algorithm is an approximate (intervals split vaguely 50-50) binary
 * search taking advantage of the fact the System.map is already sorted in
 * ascending order by the kernel makefile.  It needs to assume an average symbol
 * record length to avoid scanning the entire symbol table, but in practice the
 * search time does not seem to be especially sensitive to this choice.
 *
 * The search could be an exact binary search if the lines of System.map were
 * padded with blanks to the right.  awk '{printf "%8s%2s %-21.21s\n",$1,$2,$3}'
 * would do the trick for this but either makes the file large or truncates
 * symbols.  The approximate method seems to be plenty fast enough, costing
 * only about as much as one extra fstat() or so per process.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/utsname.h>
#include "proc/procps.h"
#include "proc/version.h"

#define MAX_ADDR_SZ 32
static char *sysmap, *sysmap_last, sysmap_fmt[10];
static int sysmap_len, sysmap_mean = 32, sysmap_addrsz;

static int three_column = 0;

/*
 * '?' is a symbol type
 * '.' is part of a name (versioning?)
 * "\t[]" are for the module name in /proc/ksyms
 */
#define LEGAL_SYSMAP_CHARS "0123456789_ ?.\n\t[]" \
                     "abcdefghijklmnopqrstuvwxyz" \
                     "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

/*******************************/

/* scan backward in a string no further than address beg looking for c */
static char * strchrrev(char *a, char *beg, char c)
{
	if (a)
		while (--a > beg && *a != c) ;
	return a;
}

/* return ptr to the beg of approximately the i-th record */
static char * addr_str(int i)
{
	char *guess = sysmap + sysmap_mean * i;
	if (!i)
		return sysmap;
	if (guess - sysmap > sysmap_len - 2)
		guess = sysmap + sysmap_len - 2;
	for (; *guess != '\n' && guess > sysmap; guess--) ;
	return guess + 1;

}

/* return ptr to symbol string (\n terminated) given beg of record ptr */
static char * sym_put(char *buf, int len, char *addrptr)
{
	char *s;
	while (*addrptr++ != ' ') ;
	if (three_column)
		while (*addrptr++ != ' ') ;
	strncpy(buf, addrptr, len);
	for (s = buf; s < buf + len; s++)
		if (*s == '\n')
			*s = '\0';
	buf[len - 1] = '\0';
	return buf;
}

/* Try to open and mmap a single symbol table file and initialize globals */
static int sysmap_mmap(const char *path, void (*message) (const char *, ...))
{
	int fd;
	struct stat sbuf;
	char *p;
	if (sysmap)		/* do nothing if already mapped */
		return 1;
	if ((fd = open(path, O_RDONLY)) < 0 || fstat(fd, &sbuf) < 0 || sbuf.st_size < MAX_ADDR_SZ + 2	/* boundary of the sanity check below */
	    || (sysmap = mmap(0, sbuf.st_size + 1,
			      PROT_READ, MAP_SHARED, fd, 0)) == (caddr_t) - 1) {
		close(fd);
		sysmap = NULL;
		return 0;
	}
	close(fd);
	sysmap_len = sbuf.st_size;
	sysmap_last = strchrrev(sysmap + sysmap_len - 2, sysmap, '\n') + 1;

	/* Now sanity check the System.map file */
	for (p = sysmap; *p != ' '
	     && ((*p >= '0' && *p <= '9') || (*p >= 'a' && *p <= 'f'))
	     && p < sysmap + MAX_ADDR_SZ; p++)	/* no-op */
		;
	three_column = (*(p + 2) == ' ');
	sysmap_addrsz = p - sysmap;
	snprintf(sysmap_fmt, sizeof sysmap_fmt, "%%0%dllx", sysmap_addrsz);
	if ((*p != ' ')
	    || (strspn(sysmap, LEGAL_SYSMAP_CHARS) != strlen(sysmap))) {
		message("Warning: %s not parseable as a System.map\n", path);
		munmap(sysmap, sysmap_len);
		sysmap = NULL;
		return 0;
	}
	/*
	 * We should make sure this file matches the kernel. We can't get a
	 * build number, but we do have a version code. The map file will
	 * contain something like "001b407c D Version_131103\n".
	 */
	do {
		char Version[32];
		sprintf(Version, " Version_%d\n", linux_version_code);
		if (strstr(sysmap, Version))
			break;
		message("Warning: %s has an incorrect kernel version.\n", path);
		munmap(sysmap, sysmap_len);
		sysmap = NULL;
		return 0;
	} while (0);
	return 1;
}

/*
 * kernel address -> name resolver.
 * returned value is only good until the next call to the function.
 */
static char * sysmap_symbol(unsigned long long address)
{
	static char rval[128], *pc, addr[MAX_ADDR_SZ];
	int i, p, n = sysmap_len / (double) sysmap_mean;

	sprintf(addr, sysmap_fmt, address);
	p = 0;
	pc = sysmap;
	while (n) {
		i = p + (n >> 1);
		if (strncmp(addr, pc = addr_str(i), sysmap_addrsz) > 0)
			p = i + 1;
		n >>= 1;
	}
	if (pc == sysmap_last)	/* scan forward but not past end */
		return sym_put(rval, sizeof rval, pc);
	while (pc <= sysmap_last && strncmp(addr, pc, sysmap_addrsz) > 0)
		pc = strchr(pc, '\n') + 1;
	if (pc == sysmap)	/* scan backward but not past beg */
		return sym_put(rval, sizeof rval, pc);
	while (pc >= sysmap && strncmp(addr, pc, sysmap_addrsz) < 0)
		pc = strchrrev(pc - 1, sysmap, '\n') + 1;
	return sym_put(rval, sizeof rval, pc);
}

static void default_message(const char *format, ...)
{
	va_list arg;

	va_start(arg, format);
	vfprintf(stderr, format, arg);
	va_end(arg);
}

int open_psdb_message(const char *override, void (*message) (const char *, ...))
{
	static char *sysmap_paths[] = {
		"/boot/System.map-%s",
		"/boot/System.map",
		"/lib/modules/%s/System.map",
		"/usr/src/linux/System.map",
		"/System.map",
		NULL
	};
	char **fmt, *env, path[64];
	struct utsname uts;

	if (getuid() != geteuid()) {
		printf("The /proc library is not secure for setuid operation.");
		exit(1);
	}
	if (getgid() != getegid()) {
		printf("The /proc library is not secure for setgid operation.");
		exit(1);
	}
	uname(&uts);
	if (override) {		/* ought to search some path */
		if (sysmap_mmap(override, message))
			return 0;
		return -1;	/* ought to return "Namelist not found." */
		/* failure is better than ignoring the user & using bad data */
	}
	/* Arrrgh, the old man page and code did not match. */
	if ((env = getenv("PS_SYSMAP")) && sysmap_mmap(env, message))
		return 0;
	if ((env = getenv("PS_SYSTEM_MAP")) && sysmap_mmap(env, message))
		return 0;
	for (fmt = sysmap_paths; *fmt; fmt++) {
		snprintf(path, sizeof path, *fmt, uts.release);
		if (sysmap_mmap(path, message))
			return 0;
	}
	return -1;
}

int open_psdb(const char *override)
{
	return open_psdb_message(override, default_message);
}

void close_psdb(void)
{
	if (sysmap)
		munmap(sysmap, sysmap_len);
	sysmap = NULL;
}

static char * find_func(unsigned long long address)
{
	if (sysmap)
		return sysmap_symbol(address);
	return NULL;
}

char * wchan(unsigned long long address)
{
	static char zero = 0;
	char *p;

	if (address) {
		p = find_func(address);
		if (!p)
			return &zero;
		if (strncmp(p, "sys_", 4) == 0)
			p += 4;
		while (*p == '_' && *p)
			++p;
	} else			/* 0 address means not in kernel space */
		p = &zero;
	return p;
}
